package main

import (
	"errors"
	"path/filepath"
	"fmt"
	"git.sr.ht/~geb/opt"
	"gocv.io/x/gocv"
	"image"
	"image/color"
	"math"
	"os"
	"os/signal"
	"strconv"
	"strings"
	"syscall"
	"time"
	ort "github.com/yalue/onnxruntime_go"
)

// FIXME The detection box doesn't keep up with movement like OpenSeeFace does.

var (
	Version string
	ModelDir = "/usr/share/noggin-models"
	ModelPackage string
)

func usage() {
	fmt.Println(`Usage: noggin [--cam=DEVICE] [--show] [--verbose] SUBSCRIPTIONS...

noggin is face tracking. It continuously prints results based on it's arguments.

--cam=DEVICE  Specify the capture device. (e.g. 0 or /dev/video0)
--show        Display a window visualizing the tracking.
--verbose     Print log messages.
--version     Print the version and exit.

Example:

    noggin position rotation --show --verbose
`)
}

func fatal(a ...any) {
	fmt.Fprintln(os.Stderr, "noggin:", fmt.Sprint(a...))
	os.Exit(1)
}

func max(a, b int) int {
	if a > b {
		return a
	}
	return b
}
func min(a, b int) int {
	if a < b {
		return a
	}
	return b
}

func lowPassFilter(v, prev, alpha float64) float64 {
	return v * alpha + (1.0 - alpha) * prev
}

func cutWord(s, word string) (string, bool) {
	if s == word {
		return "", true
	}
	if strings.HasPrefix(s, word + " ") || strings.HasPrefix(s, word + "\t") {
		return s[len(word)+1:], true
	}
	return "", false
}

type Vec struct { X, Y float32 }

// a constant for SolvePnP
var referenceContour gocv.Point3fVector = gocv.NewPoint3fVectorFromPoints([]gocv.Point3f{
	gocv.Point3f{0.45517698, 0.30089578, -0.76442945},
	gocv.Point3f{0.44899884, 0.16699584, -0.765143},
	gocv.Point3f{0.0, -0.621079, -0.28729478},
	gocv.Point3f{-0.44899884, 0.16699584, -0.765143},
	gocv.Point3f{-0.45517698, 0.30089578, -0.76442945},
	gocv.Point3f{0.0, 0.2933326, -0.1375821},
	gocv.Point3f{0.0, 0.1948287, -0.06915811},
	gocv.Point3f{0.0, 0.10384402, -0.00915182},
	gocv.Point3f{0.0, 0.0, 0.0},
	gocv.Point3f{0.08062635, -0.04127607, -0.13416104},
	gocv.Point3f{0.04643935, -0.05767522, -0.10299063},
	gocv.Point3f{0.0, -0.06875312, -0.09054535},
	gocv.Point3f{-0.04643935, -0.05767522, -0.10299063},
	gocv.Point3f{-0.08062635, -0.04127607, -0.13416104},
})

func contour(landmarks [66]Vec) gocv.Point2fVector {
	indices := []int{0, 1, 8, 15, 16, 27, 28, 29, 30, 31, 32, 33, 34, 35}
	s := make([]gocv.Point2f, len(indices))
	for i, index := range indices {
		s[i] = gocv.Point2f{landmarks[index].X, landmarks[index].Y}
	}
	return gocv.NewPoint2fVectorFromPoints(s)
}

func rollPitchYaw(rvec *gocv.Mat) (float64, float64, float64) {
	rmat := gocv.NewMat()
	defer rmat.Close()
	gocv.Rodrigues(rvec, &rmat)
	roll := math.Atan2(-float64(rmat.GetFloatAt(2, 1)), float64(rmat.GetFloatAt(2, 2)))
	pitch := math.Asin(float64(rmat.GetFloatAt(2, 0)))
	yaw := math.Atan2(-float64(rmat.GetFloatAt(1, 0)), float64(rmat.GetFloatAt(0, 0)))
	return roll*180/math.Pi, pitch*180/math.Pi, yaw*180/math.Pi
}

type BrowSub struct {}
type PositionSub struct {}
type RotationSub struct {
	Deadzone float64
	Lowpass, prevX, prevY, prevZ float64
}
type RotationChangeSub struct {
	prevRx, prevRy, prevRz float64
	Lowpass, prevX, prevY, prevZ float64
}

func main() {
	defer referenceContour.Close()
	gocv.SetNumThreads(1)

	terminate := make(chan os.Signal, 1)
	signal.Notify(terminate, os.Interrupt, syscall.SIGHUP, syscall.SIGINT, syscall.SIGTERM)

	var opts struct {
		Cam string
		Show bool
		Verbose bool
	}
	opts.Cam = "0"
	var args []string
	{
		o := opt.NewOptionSet()

		o.Func("cam", func(s string) error {
			opts.Cam = s
			return nil
		})

		o.FlagFunc("h", func() error {
			usage()
			os.Exit(0)
			panic("unreachable")
		})
		o.Alias("h", "help")

		o.BoolFunc("show", func(b bool) error {
			opts.Show = b
			return nil
		})

		o.BoolFunc("verbose", func(b bool) error {
			opts.Verbose = b
			return nil
		})

		o.FlagFunc("version", func() error {
			fmt.Println(Version)
			os.Exit(0)
			panic("unreachable")
		})

		err := o.Parse(true, os.Args[1:])
		if err != nil {
			fatal(err)
		}
		args = o.Args()
	}

	if len(args) == 0 && !opts.Show && !opts.Verbose {
		fatal("nothing was specified")
	}

	var subs []any
	{
		for _, arg := range args {
			arg := strings.Trim(arg, " \t")
			if s, ok := cutWord(arg, "brow"); ok {
				sub := &BrowSub{}
				for _, a := range strings.Fields(s) {
					fatal("unknown attribute: " + a)
				}
				subs = append(subs, sub)
			} else if s, ok := cutWord(arg, "position"); ok {
				sub := &PositionSub{}
				for _, a := range strings.Fields(s) {
					fatal("unknown attribute: " + a)
				}
				subs = append(subs, sub)
			} else if s, ok := cutWord(arg, "rotation"); ok {
				sub := &RotationSub{Lowpass: 1.0}
				for _, a := range strings.Fields(s) {
					if v, ok := strings.CutPrefix(a, "deadzone="); ok {
						f, err := strconv.ParseFloat(v, 64)
						if err == nil && f >= 0.0 {
							sub.Deadzone = f
						} else {
							fatal("invalid value: " + a)
						}
					} else if v, ok := strings.CutPrefix(a, "lowpass="); ok {
						f, err := strconv.ParseFloat(v, 64)
						if err == nil && f >= 0.0 && f <= 1.0 {
							sub.Lowpass = f
						} else {
							fatal("invalid value: " + a)
						}
					} else {
						fatal("unknown attribute: " + a)
					}
				}
				subs = append(subs, sub)
			} else if s, ok := cutWord(arg, "rotation-change"); ok {
				sub := &RotationChangeSub{Lowpass: 1.0}
				for _, a := range strings.Fields(s) {
					if v, ok := strings.CutPrefix(a, "lowpass="); ok {
						f, err := strconv.ParseFloat(v, 64)
						if err == nil && f >= 0.0 && f <= 1.0 {
							sub.Lowpass = f
						} else {
							fatal("invalid value: " + a)
						}
					} else {
						fatal("unknown attribute: " + a)
					}
				}
				subs = append(subs, sub)
			} else if arg != "" {
				fatal("unknown subscription: " + arg)
			}
		}
	}

	// TODO load normally :/
	if p := os.Getenv("NOGGIN_ONNXRUNTIME_PATH"); len(p) > 0 {
		ort.SetSharedLibraryPath(p)
	} else {
		paths, _ := filepath.Glob("/usr/lib/libonnxruntime.so*")
		if len(paths) == 0 {
			panic("no files matching: /usr/lib/libonnxruntime.so*")
		}
		ort.SetSharedLibraryPath(paths[0])
	}
	err := ort.InitializeEnvironment()
	if err != nil {
		panic(err)
	}
	defer ort.DestroyEnvironment()

	detectInput, err := ort.NewEmptyTensor[float32](ort.NewShape(1, 3, 224, 224))
	if err != nil {
		panic(err)
	}
	defer detectInput.Destroy()
	detectOutput, err := ort.NewEmptyTensor[float32](ort.NewShape(1, 2, 56, 56))
	if err != nil {
		panic(err)
	}
	defer detectOutput.Destroy()
	detectMaxpool, err := ort.NewEmptyTensor[float32](ort.NewShape(1, 2, 56, 56))
	if err != nil {
		panic(err)
	}
	defer detectMaxpool.Destroy()
	detection, err := ort.NewSession[float32](ModelDir + "/mnv3_detection_opt.onnx",
		[]string{"input"}, []string{"output", "maxpool"},
		[]*ort.Tensor[float32]{detectInput},
		[]*ort.Tensor[float32]{detectOutput, detectMaxpool})
	if err != nil {
		panic(err)
	}
	defer detection.Destroy()

	landmarkInput, err := ort.NewEmptyTensor[float32](ort.NewShape(1, 3, 224, 224))
	if err != nil {
		panic(err)
	}
	defer landmarkInput.Destroy()
	landmarkOutput, err := ort.NewEmptyTensor[float32](ort.NewShape(1, 198, 28, 28))
	if err != nil {
		panic(err)
	}
	defer landmarkOutput.Destroy()
	var landmarking *ort.Session[float32]
	{
		e := os.Getenv("NOGGIN_MODEL")
		p := e
		if p == "" {
			p = ModelDir + "/lm_model3_opt.onnx"
		}
		landmarking, err = ort.NewSession[float32](p,
			[]string{"input"}, []string{"output"},
			[]*ort.Tensor[float32]{landmarkInput},
			[]*ort.Tensor[float32]{landmarkOutput})
		if errors.Is(err, os.ErrNotExist) && e == "" && ModelPackage != "" {
			fatal("You need to install the " + ModelPackage +
				" package or set $NOGGIN_MODEL.")
		}
		if err != nil {
			if e == "" {
				panic(err)
			}
			fatal("failed to load model: " + err.Error())
		}

		if opts.Verbose {
			fmt.Fprintln(os.Stderr, "Model: " + p)
		}
	}
	defer landmarking.Destroy()

	var window *gocv.Window
	if opts.Show {
		window = gocv.NewWindow("noggin")
		defer window.Close()
	}

	cam, err := gocv.OpenVideoCaptureWithAPI(opts.Cam, gocv.VideoCaptureV4L2)
	if err != nil {
		fatal("failed to open capture device: " + err.Error())
	}
	defer cam.Close()
	cam.Set(gocv.VideoCaptureFrameWidth, 640)
	cam.Set(gocv.VideoCaptureFrameHeight, 320)
	cam.Set(gocv.VideoCaptureBufferSize, 1)

	frame := gocv.NewMat()
	defer frame.Close()

	// these are just constants for SolvePnP
	cameraMatrix := gocv.NewMatWithSize(3, 3, gocv.MatTypeCV32F)
	cameraMatrix.SetFloatAt(0, 0, float32(cam.Get(gocv.VideoCaptureFrameWidth)))
	cameraMatrix.SetFloatAt(0, 1, 0.0)
	cameraMatrix.SetFloatAt(0, 2, float32(cam.Get(gocv.VideoCaptureFrameWidth)/2))
	cameraMatrix.SetFloatAt(1, 0, 0.0)
	cameraMatrix.SetFloatAt(1, 1, float32(cam.Get(gocv.VideoCaptureFrameWidth)))
	cameraMatrix.SetFloatAt(1, 2, float32(cam.Get(gocv.VideoCaptureFrameHeight)/2))
	cameraMatrix.SetFloatAt(2, 0, 0.0)
	cameraMatrix.SetFloatAt(2, 1, 0.0)
	cameraMatrix.SetFloatAt(2, 2, 1.0)
	defer cameraMatrix.Close()
	distCoefs := gocv.NewMatWithSizeFromScalar(gocv.NewScalar(0.0, 0.0, 0.0, 0.0), 4, 1, gocv.MatTypeCV32F)
	defer distCoefs.Close()

	// these are iteratively refined by SolvePnP
	rvec := gocv.NewMatWithSize(1, 3, gocv.MatTypeCV32F)
	defer rvec.Close()
	tvec := gocv.NewMatWithSize(1, 3, gocv.MatTypeCV32F)
	defer tvec.Close()

	var detected bool
	var x0, y0, x1, y1 int  // the detection box
	var brow int
	tick := time.Tick(time.Second / 24)
	for {
		select {
		case <-terminate: return
		default:
		}

		if !frame.Empty() {
			if opts.Show {
				window.IMShow(frame)
				window.WaitKey(1)
			}
			<- tick
		}

		if ok := cam.Read(&frame); !ok {
			if opts.Verbose {
				fmt.Fprintln(os.Stderr, "Capture device closed.")
			}
			return
		}
		if frame.Empty() {
			if opts.Verbose {
				fmt.Fprintln(os.Stderr, "Skipping empty frame.")
			}
			continue
		}

		if !detected {
			x0, y0, x1, y1, detected = detect(frame,
			func(input []float32) []float32 {
				copy(detectInput.GetData(), input)
				if err := detection.Run(); err != nil {
					panic(err)
				}
				return append(detectOutput.GetData(), detectMaxpool.GetData()...)
			})

			if detected && opts.Verbose {
				fmt.Fprintln(os.Stderr, "Detected.")
			}
		}
		if !detected {
			continue
		}

		x0 = max(0, x0 - int(float64(x1 - x0) * 0.1))
		y0 = max(0, y0 - int(float64(y1 - y0) * 0.125))
		x1 = min(x1 + int(float64(x1 - x0) * 0.1), frame.Cols()-1)
		y1 = min(y1 + int(float64(y1 - y0) * 0.125), frame.Rows()-1)

		if x1 - x0 < 4 || y1 - y0 < 4 {
			detected = false
			if opts.Verbose {
				fmt.Fprintln(os.Stderr,
					"Lost detection (detection box too small).")
			}
			continue
		}

		if opts.Show {
			gocv.Rectangle(&frame, image.Rect(x0, y0, x1, y1),
				color.RGBA{0, 0, 255, 0}, 2)
		}

		lms, conf := landmark(frame, x0, y0, x1, y1,
		func(input []float32) []float32 {
			copy(landmarkInput.GetData(), input)
			if err := landmarking.Run(); err != nil {
				panic(err)
			}
			return landmarkOutput.GetData()
		})

		if conf >= 0.6 {
			// the +/-5 help to compensate for the movement FIXME
			x0, y0 = int(lms[0].X)-5, int(lms[0].Y)
			x1, y1 = int(lms[1].X)+5, int(lms[1].Y)
			for _, lm := range lms {
				x0 = min(x0, int(lm.X)-5)
				y0 = min(y0, int(lm.Y))
				x1 = max(x1, int(lm.X)+5)
				y1 = max(y1, int(lm.Y))
			}
			if opts.Show {
				for _, lm := range lms {
					gocv.Circle(&frame, image.Point{int(lm.X), int(lm.Y)}, 1,
						color.RGBA{0, 255, 0, 0}, -1)
				}
			}
		} else {
			if opts.Verbose {
				fmt.Fprintln(os.Stderr,
					"Lost detection (landmark confidence too low).")
			}
			detected = false
			continue
		}

		var rx, ry, rz float64
		LOOP:
		for _, sub := range subs {
			switch sub.(type) {
			case *PositionSub, *RotationSub, *RotationChangeSub:
				cont := contour(lms)
				ok := gocv.SolvePnP(referenceContour, cont, &cameraMatrix,
					&distCoefs, &rvec, &tvec, true, 0)
				if ok {
					rx, ry, rz = rollPitchYaw(&rvec)
				} else if opts.Verbose {
					fmt.Fprintln(os.Stderr, "Failed to solve head points.")
				}
				cont.Close()
				break LOOP
			}
		}

		for _, sub := range subs {
			switch sub := sub.(type) {
			case *BrowSub:
				nose := lms[27].Y - lms[30].Y
				left := ((lms[22].Y + lms[26].Y) / 2.0 - lms[27].Y) / nose
				right := ((lms[17].Y + lms[21].Y) / 2.0 - lms[27].Y) / nose
				b := (left + right) / 2.0
				if brow >= 0 {
					if rx < -10.0 {
						if b > 0.55 { brow = 1 }
						if b < 0.52 { brow = 0 }
					} else if (rx > 16.0 && math.Abs(ry) > 16.0) || (rx > 20.0 && math.Abs(ry) > 10.0) {
						if b > 0.61 { brow = 1 }
						if b < 0.58 { brow = 0 }
					} else {
						if b > 0.57 { brow = 1 }
						if b < 0.52 { brow = 0 }
					}
				}
				if brow <= 0 {
					if rx < -10.0 {
						if b < 0.38 { brow = -1 }
						if b > 0.43 { brow = 0 }
					} else if (rx > 16.0 && math.Abs(ry) > 16.0) || (rx > 20.0 && math.Abs(ry) > 10.0) {
						if b < 0.39 { brow = -1 }
						if b > 0.47 { brow = 0 }
					} else {
						if b < 0.39 { brow = -1 }
						if b > 0.47 { brow = 0 }
					}
				}
				if brow > 0 {
					fmt.Println("brow: up")
				} else if brow < 0 {
					fmt.Println("brow: down")
				} else {
					fmt.Println("brow: neutral")
				}
			case *PositionSub:
				x := float64(tvec.GetFloatAt(0, 0))
				y := float64(tvec.GetFloatAt(0, 1))
				z := float64(tvec.GetFloatAt(0, 2))
				fmt.Println("position:", x, y, z)
			case *RotationSub:
				h := math.Hypot(rx, ry)
				s := math.Max(0.0, h - sub.Deadzone)
				x := lowPassFilter(rx * s / h, sub.prevX, sub.Lowpass)
				y := lowPassFilter(ry * s / h, sub.prevY, sub.Lowpass)
				z := lowPassFilter(rz * s / h, sub.prevZ, sub.Lowpass)
				fmt.Println("rotation:", x, y, z)
				sub.prevX, sub.prevY, sub.prevZ = x, y, z
			case *RotationChangeSub:
				x := lowPassFilter((rx - sub.prevRx), sub.prevX, sub.Lowpass)
				y := lowPassFilter((ry - sub.prevRy), sub.prevY, sub.Lowpass)
				z := lowPassFilter((rz - sub.prevRz), sub.prevZ, sub.Lowpass)
				fmt.Println("rotation-change:", x, y, z)
				sub.prevRx, sub.prevRy, sub.prevRz = rx, ry, rz
				sub.prevX, sub.prevY, sub.prevZ = x, y, z
			}
		}
	}
}

func detect(img gocv.Mat, compute func([]float32) []float32) (int, int, int, int, bool) {
	img = img.Clone()
	defer img.Close()

	img.ConvertTo(&img, gocv.MatTypeCV32F)
	scale := gocv.NewMatWithSizeFromScalar(
		gocv.NewScalar(0.01742919, 0.017507, 0.01712475, 1.0),
		img.Rows(), img.Cols(), img.Type())
	defer scale.Close()
	gocv.Multiply(img, scale, &img)
	blob := gocv.BlobFromImage(img, 1.0, image.Pt(224, 224),
		gocv.NewScalar(2.117904, 2.0357141, 1.8044444, 0.0), true, false)
	defer blob.Close()

	d, err := blob.DataPtrFloat32()
	if err != nil {
		panic(err)
	}
	c := compute(d)

	// a heatmap, a radius map and the heatmap after maxpooling
	// 3136 is the number of heatmap pixels (56x56)
	heatmap, radmap, maxpool := c[:3136], c[3136:2*3136], c[2*3136:]

	hottest := 0
	for i := range heatmap {
		if heatmap[i] == maxpool[i] && heatmap[i] > heatmap[hottest] {
			hottest = i
		}
	}
	if heatmap[hottest] < 0.6 {
		return 0, 0, 0, 0, false
	}

	radius := int(radmap[hottest] * 112)
	x := (hottest % 56) * 4
	y := (hottest / 56) * 4
	x0 := (x - radius) * img.Cols() / 224
	y0 := (y - radius) * img.Rows() / 224
	x1 := (x + radius) * img.Cols() / 224
	y1 := (y + radius) * img.Rows() / 224
	return x0, y0, x1, y1, true
}

// https://github.com/emilianavt/OpenSeeFace/issues/24
// "the inverse of multiplying by 16 and taking the sigmoid"
func logit(x float32) float32 {
	if x <= 0.0 {
		x = 0.0000001
	} else if x >= 1.0 {
		x = 0.9999999
	}
	x /= 1.0 - x
	return float32(math.Log(float64(x)) / 16.0)
}

func landmark(img gocv.Mat, x0, y0, x1, y1 int, compute func([]float32) []float32) ([66]Vec, float32) {
	crop := img.Region(image.Rect(x0, y0, x1, y1))
	defer crop.Close()
	crop = crop.Clone()
	defer crop.Close()
	crop.ConvertTo(&crop, gocv.MatTypeCV32F)
	scale := gocv.NewMatWithSizeFromScalar(
		gocv.NewScalar(0.01742919, 0.017507, 0.01712475, 1.0),
		crop.Rows(), crop.Cols(), crop.Type())
	defer scale.Close()
	gocv.Multiply(crop, scale, &crop)
	blob := gocv.BlobFromImage(crop, 1.0, image.Pt(224, 224),
		gocv.NewScalar(2.117904, 2.0357141, 1.8044444, 0.0), true, false)
	defer blob.Close()

	d, err := blob.DataPtrFloat32()
	if err != nil {
		panic(err)
	}
	c := compute(d)

	// a heatmap and two more for extra precision, per landmark
	// 66 is the number of landmarks
	// 784 is the number of heatmap pixels (28x28)
	heatmaps := c[:66*784]
	yHonemaps, xHonemaps := c[66*784:2*66*784], c[2*66*784:]

	xScale, yScale := float32(x1 - x0) / 224.0, float32(y1 - y0) / 224.0
	var landmarks [66]Vec
	var confidence float32
	for i := range landmarks {
		hMap := heatmaps[i*784:i*784+784]
		xMap, yMap := xHonemaps[i*784:i*784+784], yHonemaps[i*784:i*784+784]
		hottest := 0
		for j := range hMap {
			if hMap[j] > hMap[hottest] {
				hottest = j
			}
		}
		confidence += hMap[hottest]
		xHone := 223.0 * logit(xMap[hottest])
		yHone := 223.0 * logit(yMap[hottest])
		x := float32(x0) + xScale * (float32(hottest % 28) * 223.0/27.0 + xHone)
		y := float32(y0) + yScale * (float32(hottest / 28) * 223.0/27.0 + yHone)
		landmarks[i] = Vec{x, y}
	}
	confidence /= 66.0

	return landmarks, confidence
}
